@@ -1,7 +1,7 @@
 /*
  * signal.c - signal handling
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2006 Shiro Kawai, All rights reserved.
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: signal.c,v 1.29 2004/07/15 07:10:07 shirok Exp $
+ *  $Id: signal.c,v 1.41 2006/11/04 09:56:59 shirok Exp $
  */
 
 #include <stdlib.h>
@@ -92,9 +92,11 @@ static struct sigHandlersRec {
 /* Table of signals and its initial behavior. */
 #define SIGDEF_NOHANDLE 0       /* Gauche doesn't install a signal handler,
                                    leaving it to the application. */
-#define SIGDEF_ERROR    1       /* Gauche installs a default signal handler
+#define SIGDEF_DFL      1       /* Gauche resets the singal handler to
+                                   SIG_DFL. */
+#define SIGDEF_ERROR    2       /* Gauche installs a default signal handler
                                    that raises an error. */
-#define SIGDEF_EXIT     2       /* Gauche installs a handler that calls
+#define SIGDEF_EXIT     3       /* Gauche installs a handler that calls
                                    Scm_Exit(). */
 
 #define SIGDEF(x, flag)  { #x, x, flag }
@@ -144,7 +146,7 @@ static struct sigdesc {
     SIGDEF(SIGSTKFLT, SIGDEF_ERROR),  /* Stack fault */
 #endif
 #ifdef SIGCHLD
-    SIGDEF(SIGCHLD, SIGDEF_NOHANDLE), /* Child status has changed (POSIX) */
+    SIGDEF(SIGCHLD, SIGDEF_DFL),      /* Child status has changed (POSIX) */
 #endif
 #ifdef SIGCONT
     SIGDEF(SIGCONT, SIGDEF_NOHANDLE), /* Continue (POSIX) */
@@ -259,15 +261,21 @@ static ScmObj default_sighandler(ScmObj *args, int nargs, void *data)
         }
     }
     if (name) {
-        Scm_Error("unhandled signal %d (%s)", signum, name);
+        Scm_RaiseCondition(SCM_OBJ(SCM_CLASS_UNHANDLED_SIGNAL_ERROR),
+                           "signal", SCM_MAKE_INT(signum),
+                           SCM_RAISE_CONDITION_MESSAGE,
+                           "unhandled signal %d (%s)", signum, name);
     } else {
-        Scm_Error("unhandled signal %d (unknown signal)", signum);
+        Scm_RaiseCondition(SCM_OBJ(SCM_CLASS_UNHANDLED_SIGNAL_ERROR),
+                           "signal", SCM_MAKE_INT(signum),
+                           SCM_RAISE_CONDITION_MESSAGE,
+                           "unhandled signal %d (unknown signal)", signum);
     }
     return SCM_UNDEFINED;       /* dummy */
 }
 
 static SCM_DEFINE_STRING_CONST(default_sighandler_name,
-                               "%default-signal-handler", 24, 24);
+                               "%default-signal-handler", 23, 23);
 static SCM_DEFINE_SUBR(default_sighandler_stub, 1, 0,
                        SCM_OBJ(&default_sighandler_name),
                        default_sighandler,
@@ -291,6 +299,23 @@ static SCM_DEFINE_SUBR(exit_sighandler_stub, 1, 0,
 
 #define EXIT_SIGHANDLER    SCM_OBJ(&exit_sighandler_stub)
 
+#if 0                           /* not used for now */
+/* For some signals, gauche does nothing */
+static ScmObj through_sighandler(ScmObj *args, int nargs, void *data)
+{
+    return SCM_UNDEFINED;
+}
+
+static SCM_DEFINE_STRING_CONST(through_sighandler_name,
+                               "%through-signal-handler", 20, 20);
+static SCM_DEFINE_SUBR(through_sighandler_stub, 1, 0,
+                       SCM_OBJ(&through_sighandler_name),
+                       through_sighandler,
+                       NULL, NULL);
+
+#define THROUGH_SIGHANDLER    SCM_OBJ(&through_sighandler_stub)
+#endif
+
 /*
  * sigset class
  */
@@ -361,6 +386,10 @@ ScmObj Scm_SysSigsetFill(ScmSysSigset *set, int emptyp)
     return SCM_OBJ(set);
 }
 
+/*
+ * Signal handling
+ */
+
 /*
  * System's signal handler - just enqueue the signal
  */
@@ -372,19 +401,37 @@ static void sig_handle(int signum)
        terminating and in the cleanup phase. */
     if (vm == NULL) return;
     
-    if (vm->sigOverflow) return;
-    
-    if (vm->sigQueueHead <= vm->sigQueueTail) {
-        vm->sigQueue[vm->sigQueueTail++] = signum;
-        if (vm->sigQueueTail >= SCM_VM_SIGQ_SIZE) {
-            vm->sigQueueTail = 0;
+    if (vm->sigq.overflow) {
+        /* TODO: we should have better fallback strategy. */
+        struct sigdesc *desc = sigDesc;
+        const char *signame = "unknown signal";
+        for (; desc->name; desc++) {
+            if (desc->num == signum) {
+                signame = desc->name;
+                break;
+            }
         }
-    } else {
-        vm->sigQueue[vm->sigQueueTail++] = signum;
+        Scm_Warn("Signal queue overflow by %s(%d)\n", signame, signum);
+        return;
     }
-    if (vm->sigQueueTail == vm->sigQueueHead) {
-        Scm_Error("signal queue overflow\n");
+    vm->sigq.queue[vm->sigq.tail++] = signum;
+    if (vm->sigq.tail >= SCM_VM_SIGQ_SIZE) {
+        vm->sigq.tail = 0;
     }
+    if (vm->sigq.tail == vm->sigq.head) vm->sigq.overflow++;
+    vm->queueNotEmpty |= SCM_VM_SIGQ_MASK;
+}
+
+/*
+ * Initializes signal queue
+ */
+void Scm_SignalQueueInit(ScmSignalQueue* q)
+{
+    int i;
+    for (i=0; i<SCM_VM_SIGQ_SIZE; i++) q->queue[i] = -1;
+    q->head = q->tail = 0;
+    q->overflow = 0;
+    q->pending = SCM_NIL;
 }
 
 /*
@@ -394,6 +441,7 @@ static void sig_handle(int signum)
 void Scm_SigCheck(ScmVM *vm)
 {
     ScmObj tail, cell, sp;
+    ScmSignalQueue *q = &vm->sigq;
     sigset_t omask;
     int sigQsize, i;
     int sigQcopy[SCM_VM_SIGQ_SIZE]; /* copy of signal queue */
@@ -401,17 +449,18 @@ void Scm_SigCheck(ScmVM *vm)
     /* Copy VM's signal queue to local storage, for we can't call
        storage allocation during blocking signals. */
     SIGPROCMASK(SIG_BLOCK, &sigHandlers.masterSigset, &omask);
-    for (sigQsize = 0; vm->sigQueueHead != vm->sigQueueTail; sigQsize++) {
-        sigQcopy[sigQsize] = vm->sigQueue[vm->sigQueueHead++];
-        if (vm->sigQueueHead >= SCM_VM_SIGQ_SIZE) vm->sigQueueHead = 0;
+    for (sigQsize = 0; q->head != q->tail; sigQsize++) {
+        sigQcopy[sigQsize] = q->queue[q->head++];
+        if (q->head >= SCM_VM_SIGQ_SIZE) q->head = 0;
     }
-    vm->sigOverflow = 0; /*TODO: we should do something*/
+    q->overflow = 0; /*TODO: we should do something*/
+    vm->queueNotEmpty &= ~SCM_VM_SIGQ_MASK;
     SIGPROCMASK(SIG_SETMASK, &omask, NULL);
 
     /* Now, prepare queued signal handlers
        If an error is thrown in this loop, the queued signals will be
        lost---it doesn't look like so, but I may overlook something. */
-    tail = vm->sigPending;
+    tail = q->pending;
     if (!SCM_NULLP(tail)) tail = Scm_LastPair(tail);
     for (i=0; i<sigQsize; i++) {
         if (SCM_PROCEDUREP(sigHandlers.handlers[sigQcopy[i]])) {
@@ -419,7 +468,7 @@ void Scm_SigCheck(ScmVM *vm)
                              SCM_MAKE_INT(sigQcopy[i]),
                              SCM_NIL);
             if (SCM_NULLP(tail)) {
-                vm->sigPending = tail = cell;
+                q->pending = tail = cell;
             } else {
                 SCM_SET_CDR(tail, cell);
                 tail = SCM_CDR(tail);
@@ -431,10 +480,10 @@ void Scm_SigCheck(ScmVM *vm)
        of those handlers, the rest of handlers remain in the queue. */
     /* TODO: if VM is active, it'd be better to make the active VM to handle
        those handler procs, instead of calling Scm_Eval. */
-    SCM_FOR_EACH(sp, vm->sigPending) {
+    SCM_FOR_EACH(sp, q->pending) {
         ScmObj h = SCM_CAR(sp);
-        vm->sigPending = SCM_CDR(sp);
-        Scm_Apply(SCM_CAR(h), SCM_LIST1(SCM_CDR(h)));
+        q->pending = SCM_CDR(sp);
+        Scm_ApplyRec(SCM_CAR(h), SCM_LIST1(SCM_CDR(h)));
     }
 }
 
@@ -567,14 +616,24 @@ void Scm_SetMasterSigmask(sigset_t *set)
         } else if (!sigismember(&sigHandlers.masterSigset, desc->num)
                    && sigismember(set, desc->num)) {
             /* add sighandler if necessary */
-            if (desc->defaultHandle != SIGDEF_NOHANDLE) {
+            if (desc->defaultHandle == SIGDEF_DFL) {
+                if (sigaction(desc->num, &actoff, NULL) != 0) {
+                    Scm_SysError("sigaction on %d failed", desc->num);
+                }
+                sigHandlers.handlers[desc->num] = SCM_TRUE;
+            } else if (desc->defaultHandle != SIGDEF_NOHANDLE) {
                 if (sigaction(desc->num, &acton, NULL) != 0) {
                     Scm_SysError("sigaction on %d failed", desc->num);
                 }
-                if (desc->defaultHandle == SIGDEF_EXIT) {
-                    sigHandlers.handlers[desc->num] = EXIT_SIGHANDLER;
-                } else {
+                switch (desc->defaultHandle) {
+                case SIGDEF_ERROR:
                     sigHandlers.handlers[desc->num] = DEFAULT_SIGHANDLER;
+                    break;
+                case SIGDEF_EXIT:
+                    sigHandlers.handlers[desc->num] = EXIT_SIGHANDLER;
+                    break;
+                default:
+                    Scm_Panic("Scm_SetMasterSigmask: can't be here");
                 }
             }
         }
@@ -590,15 +649,42 @@ void Scm_SetMasterSigmask(sigset_t *set)
 ScmObj Scm_SysSigmask(int how, ScmSysSigset *newmask)
 {
     ScmSysSigset *oldmask = make_sigset();
-    if (how != SIG_SETMASK && how != SIG_BLOCK && how != SIG_UNBLOCK) {
-        Scm_Error("bad 'how' argument for signal mask action: %d", how);
+    sigset_t *newset = NULL;
+
+    if (newmask) {
+        newset = &(newmask->set);
+        if (how != SIG_SETMASK && how != SIG_BLOCK && how != SIG_UNBLOCK) {
+            Scm_Error("bad 'how' argument for signal mask action: %d", how);
+        }
     }
-    if (SIGPROCMASK(how, &(newmask->set), &(oldmask->set)) != 0) {
+    if (SIGPROCMASK(how, newset, &(oldmask->set)) != 0) {
         Scm_Error("sigprocmask failed");
     }
     return SCM_OBJ(oldmask);
 }
 
+/*
+ * Reset signal handlers except the masked ones.
+ * This is called just before we change the signal mask and call exec(2),
+ * so that we can avoid the hazard that the signal handler is called
+ * between sigsetmask and exec.
+ */
+void Scm_ResetSignalHandlers(sigset_t *mask)
+{
+    struct sigdesc *desc = sigDesc;
+    struct sigaction act;
+
+    for (; desc->name; desc++) {
+        if (!sigismember(&sigHandlers.masterSigset, desc->num)
+            && (!mask || !sigismember(mask, desc->num))) {
+            act.sa_flags = 0;
+            act.sa_handler = SIG_IGN;
+            // NB: we tolerate failure of this
+            sigaction(desc->num, &act, NULL);
+        }
+    }
+}
+
 /*
  * sigsuspend
  */
@@ -609,7 +695,7 @@ static void scm_sigsuspend(sigset_t *mask)
     ScmVM *vm = Scm_VM();
     for (;;) {
         SIGPROCMASK(SIG_BLOCK, &sigHandlers.masterSigset, &omask);
-        if (vm->sigQueueTail != vm->sigQueueHead) {
+        if (vm->sigq.tail != vm->sigq.head) {
             SIGPROCMASK(SIG_SETMASK, &omask, NULL);
             Scm_SigCheck(vm);
             continue;
@@ -680,8 +766,8 @@ void Scm__InitSignal(void)
     sigemptyset(&sigHandlers.masterSigset);
     for (i=0; i<NSIG; i++) sigHandlers.handlers[i] = SCM_FALSE;
     
-    Scm_InitBuiltinClass(&Scm_SysSigsetClass, "<sys-sigset>", NULL,
-                         sizeof(ScmSysSigset), mod);
+    Scm_InitStaticClass(&Scm_SysSigsetClass, "<sys-sigset>",
+                        mod, NULL, 0);
 
     for (desc = sigDesc; desc->name; desc++) {
         SCM_DEFINE(mod, desc->name, SCM_MAKE_INT(desc->num));
