@@ -1,7 +1,7 @@
 /*
  * system.c - system interface
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2006 Shiro Kawai, All rights reserved.
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: system.c,v 1.60 2004/07/27 10:23:47 shirok Exp $
+ *  $Id: system.c,v 1.84 2006/11/11 07:48:52 shirok Exp $
  */
 
 #include <stdio.h>
@@ -43,6 +43,7 @@
 #include <string.h>
 #include <sys/types.h>
 #include <sys/stat.h>
+#include <fcntl.h>
 #ifndef __MINGW32__
 #include <grp.h>
 #include <pwd.h>
@@ -50,17 +51,20 @@
 #include <windows.h>
 #include <lm.h>
 #include <tlhelp32.h>
-#include <fcntl.h>
 #endif  /*__MINGW32__*/
 
 #define LIBGAUCHE_BODY
 #include "gauche.h"
 #include "gauche/class.h"
+#include "gauche/bignum.h"
 #include "gauche/builtin-syms.h"
 
 #ifdef HAVE_GLOB_H
 #include <glob.h>
 #endif
+#ifdef HAVE_SCHED_H
+#include <sched.h>
+#endif
 
 /*
  * Auxiliary system interface functions.   See syslib.stub for
@@ -183,17 +187,19 @@ ScmObj Scm_ReadDirectory(ScmString *pathname)
 }
 
 /* Glob()function. */
-/* TODO: allow to take optional flags */
+/* NB: This will soon go away.  sys-glob will be implemented in Scheme
+   on top of sys-readdir and some text manipulation, so that we'll be
+   free from system-dependent gotchas. */
 ScmObj Scm_GlobDirectory(ScmString *pattern)
 {
-#ifdef HAVE_GLOB_H
+#if defined(HAVE_GLOB_H)
     glob_t globbed;
     ScmObj head = SCM_NIL, tail = SCM_NIL;
     int i, r;
     SCM_SYSCALL(r, glob(Scm_GetStringConst(pattern), 0, NULL, &globbed));
     if (r) {
         globfree(&globbed);
-#ifdef GLOB_NOMATCH
+#if defined(GLOB_NOMATCH)
         if (r == GLOB_NOMATCH) return SCM_NIL;
 #endif /*!GLOB_NOMATCH*/
         Scm_Error("Couldn't glob %S", pattern);
@@ -204,14 +210,44 @@ ScmObj Scm_GlobDirectory(ScmString *pattern)
     }
     globfree(&globbed);
     return head;
-#else  /*!HAVE_GLOB_H*/
+#elif defined(__MINGW32__)
+    /* We provide alternative using Windows API */
+    HANDLE dirp;
+    WIN32_FIND_DATA fdata;
+    DWORD winerrno;
+    const char *path = Scm_GetStringConst(pattern);
+    ScmObj head = SCM_NIL, tail = SCM_NIL;
+    
+    dirp = FindFirstFile(path, &fdata);
+    if (dirp == INVALID_HANDLE_VALUE) {
+	if ((winerrno = GetLastError()) != ERROR_FILE_NOT_FOUND) goto err;
+	return head;
+    }
+    SCM_APPEND1(head, tail, SCM_MAKE_STR_COPYING(fdata.cFileName));
+    while (FindNextFile(dirp, &fdata) != 0) {
+	SCM_APPEND1(head, tail, SCM_MAKE_STR_COPYING(fdata.cFileName));
+    }
+    winerrno = GetLastError();
+    FindClose(dirp);
+    if (winerrno != ERROR_NO_MORE_FILES) goto err;
+    return head;
+ err:
+    Scm_Error("Searching directory failed by windows error %d",
+	      winerrno);
+    return SCM_UNDEFINED;	/* dummy */
+#else  /*!HAVE_GLOB_H && !__MINGW32__*/
     Scm_Error("glob-directory is not supported on this architecture.");
     return SCM_UNDEFINED;
-#endif /*!HAVE_GLOB_H*/
+#endif /*!HAVE_GLOB_H && !__MINGW32__*/
 }
 
-/*
+/*===============================================================
  * Pathname manipulation
+ *
+ *  It gets complicated since the byte '/' and '\\' can appear in
+ *  the trailing octets of a multibyte character.
+ *  Assuming these operations won't be a bottleneck, we use simple and
+ *  straightforward code rather than tricky and fast one.
  */
 
 /* Returns the system's native pathname delimiter. */
@@ -224,206 +260,400 @@ const char *Scm_PathDelimiter(void)
 #endif /*__MINGW32__*/
 }
 
-ScmObj Scm_NormalizePathname(ScmString *pathname, int flags)
-{
-    const char *str = SCM_STRING_START(pathname), *srcp = str;
-    int size = SCM_STRING_SIZE(pathname);
-    char *buf = NULL, *dstp;
-    int bottomp = FALSE;
-    
-#define SKIP_SLASH \
-    while ((*srcp == '/' || *srcp == '\\') && srcp < str+size) { srcp++; }
+#ifdef __MINGW32__
+#define SEPARATOR '\\'
+#define ROOTDIR   "\\"
+#else
+#define SEPARATOR '/'
+#define ROOTDIR   "/"
+#endif
 
-    if ((flags & SCM_PATH_EXPAND) && size >= 1 && *str == '~') {
-#ifndef __MINGW32__
-        /* ~user magic */
-        const char *p = str+1;
-        struct passwd *pwd;
-        int dirlen;
-        
-        for (; p < str+size && *p != '/' && *p != '\0'; p++)
-            ;
-        if (p == str+1) {
-            pwd = getpwuid(geteuid());
-            if (pwd == NULL) {
-                Scm_SigCheck(Scm_VM());
-                Scm_SysError("couldn't get home directory.\n");
-            }
-        } else {
-            char *user = (char *)SCM_MALLOC_ATOMIC(p-str);
-            memcpy(user, str+1, p-str-1);
-            user[p-str-1] = '\0';
-            pwd = getpwnam(user);
-            if (pwd == NULL) {
-                Scm_SigCheck(Scm_VM());
-                Scm_Error("couldn't get home directory of user \"%s\".\n",
-                          user);
-            }
-        }
-        srcp = p;
-        SKIP_SLASH;
-        dirlen = strlen(pwd->pw_dir);
-        buf = SCM_NEW_ATOMIC2(char*, dirlen+size+1);
-        strcpy(buf, pwd->pw_dir);
-        dstp = buf + dirlen;
-        if (*(dstp-1) != '/') { *dstp++ = '/'; *(dstp+1) = '\0'; }
-#else  /* __MINGW32__ */
-	/* we don't support '~' magic on win32 native */
-	buf = SCM_NEW_ATOMIC2(char*, size+1);
-	dstp = buf;
-#endif /* __MINGW32__ */
-    } else if ((flags & SCM_PATH_ABSOLUTE) && *str != '/') {
-        int dirlen;
-#define GETCWD_PATH_MAX 1024  /* TODO: must be configured */
-        char p[GETCWD_PATH_MAX];
-        if (getcwd(p, GETCWD_PATH_MAX-1) == NULL) {
+/* Returns the pointer to the first path separator character,
+   or NULL if no separator is found. */
+static const char *get_first_separator(const char *path, const char *end)
+{
+    const char *p = path;
+    while (p < end) {
+        if (*p == '/' || *p == '\\') return p;
+        p += SCM_CHAR_NFOLLOWS(*p)+1;
+    }
+    return NULL;
+}
+
+/* Returns the pointer to the last path separator character,
+   or NULL if no separator is found. */
+static const char *get_last_separator(const char *path, const char *end)
+{
+    const char *p = path, *last = NULL;
+    while (p < end) {
+        if (*p == '/' || *p == '\\') last = p;
+        p += SCM_CHAR_NFOLLOWS(*p)+1;
+    }
+    return last;
+}
+
+static const char *skip_separators(const char *p, const char *end)
+{
+    while (p < end) {
+        if (*p != '/' && *p != '\\') break;
+        p += SCM_CHAR_NFOLLOWS(*p)+1;
+    }
+    return p;
+}
+
+/* Returns the end pointer sans trailing separators. */
+static const char *truncate_trailing_separators(const char *path,
+                                                const char *end)
+{
+    const char *p = get_first_separator(path, end), *q;
+    if (p == NULL) return end;
+    for (;;) {
+        q = skip_separators(p, end);
+        if (q == end) return p;
+        p = get_first_separator(q, end);
+        if (p == NULL) return end;
+    }
+}
+
+/* A utility for tilde expansion.  They are called only on
+   Unix variants, so we only need to check '/'. */
+static void put_user_home(ScmDString *dst,
+                          const char *name,
+                          const char *end)
+{
+    struct passwd *pwd;
+    int dirlen;
+
+    if (name == end) {
+        pwd = getpwuid(geteuid());
+        if (pwd == NULL) {
             Scm_SigCheck(Scm_VM());
-            Scm_SysError("couldn't get current directory.");
-        }
-        dirlen = strlen(p);
-        buf = SCM_NEW_ATOMIC2(char*, dirlen+size+1);
-        strcpy(buf, p);
-        dstp = buf + dirlen;
-        if (*(dstp-1) != '/') *dstp++ = '/';
-    } else if (flags & SCM_PATH_CANONICALIZE) {
-        dstp = buf = SCM_NEW_ATOMIC2(char*, size+1);
-        if (*str == '/') {
-            *dstp++ = '/';
-            SKIP_SLASH;
+            Scm_SysError("couldn't get home directory.\n");
         }
     } else {
-        return SCM_OBJ(pathname); /* nothing to do */
+        int namesiz = end - name;
+        char *uname = (char*)SCM_MALLOC_ATOMIC(namesiz+1);
+        memcpy(uname, name, namesiz);
+        uname[namesiz] = '\0';
+        pwd = getpwnam(uname);
+        if (pwd == NULL) {
+            Scm_SigCheck(Scm_VM());
+            Scm_Error("couldn't get home directory of user \"%s\".\n", uname);
+        }
     }
+    dirlen = strlen(pwd->pw_dir);
+    Scm_DStringPutz(dst, pwd->pw_dir, dirlen);
+    if (pwd->pw_dir[dirlen-1] != '/') Scm_DStringPutc(dst, '/');
+}
 
+/* SRC points to the pathname string beginning with '~'.  Expand it
+   to the user's home directory, leaving the partial result in DST.
+   Returns the pointer into SRC sans tilde prefix (e.g. removed "~user/").
+   The returned pointer may points just past the last char of SRC. */
+static const char *expand_tilde(ScmDString *dst,
+                                const char *src,
+                                const char *end)
+{
+    const char *sep = get_first_separator(src, end);
+
+    if (sep == NULL) {
+        put_user_home(dst, src+1, end);
+        return end;
+    } else {
+        put_user_home(dst, src+1, sep);
+        return skip_separators(sep, end);
+    }
+}
+
+/* Put current dir to DST */
+static void put_current_dir(ScmDString *dst)
+{
+    int dirlen;
+#define GETCWD_PATH_MAX 1024  /* TODO: must be configured */
+    char p[GETCWD_PATH_MAX];
+    if (getcwd(p, GETCWD_PATH_MAX-1) == NULL) {
+        Scm_SigCheck(Scm_VM());
+        Scm_SysError("couldn't get current directory.");
+    }
+    dirlen = strlen(p);
+    Scm_DStringPutz(dst, p, dirlen);
+    if (p[dirlen-1] != '/' && p[dirlen-1] != '\\') {
+        Scm_DStringPutc(dst, SEPARATOR);
+    }
+#undef GETCWD_PATH_MAX
+}
+
+#if defined(__MINGW32__)
+/* win32 specific; copy pathname with replacing '/' by '\\'. */
+static void copy_win32_path(ScmDString *dst,
+                            const char *srcp,
+                            const char *end)
+{
+    while (srcp < end) {
+        ScmChar ch;
+        if (*srcp == '/' || *srcp == '\\') {
+            Scm_DStringPutc(dst, SEPARATOR);
+        } else {
+            SCM_CHAR_GET(srcp, ch);
+            Scm_DStringPutc(dst, ch);
+        }
+        srcp += SCM_CHAR_NBYTES(ch);
+    }
+}
+#endif /* __MINGW32__ */
+
+ScmObj Scm_NormalizePathname(ScmString *pathname, int flags)
+{
+    u_int size;
+    const char *str = Scm_GetStringContent(pathname, &size, NULL, NULL);
+    const char *srcp = str;
+    const char *endp = str + size;
+    ScmDString buf;
+
+    Scm_DStringInit(&buf);
+
+    /* Preprocess.  We expand tilde (on unix platform), and prepend the
+       current directory to the relative pathname if absolutize is required.
+       For canonicalization, we also put any absolute prefix into buf, so
+       that srcp points to the relative path part after this. */
+#if !defined(__MINGW32__)
+    if ((flags & SCM_PATH_EXPAND) && size >= 1 && *str == '~') {
+        srcp = expand_tilde(&buf, srcp, endp);
+    } else if (endp > srcp && *srcp == '/') {
+        /* Path is absolute */
+        if (flags & SCM_PATH_CANONICALIZE) {
+            Scm_DStringPutc(&buf, SEPARATOR);
+            srcp = skip_separators(srcp, endp);
+        }
+    } else {
+        /* Path is relative */
+        if (flags & SCM_PATH_ABSOLUTE) {
+            put_current_dir(&buf);
+        }
+    }
     if (!(flags & SCM_PATH_CANONICALIZE)) {
-        size -= srcp-str;
-        memcpy(dstp, srcp, size);
-        *(dstp + size) = '\0';
-        return Scm_MakeString(buf, (dstp-buf)+size, -1, SCM_MAKSTR_COPYING);
+        Scm_DStringPutz(&buf, srcp, endp - srcp);
+	return Scm_DStringGet(&buf, 0);
     }
+#else /* __MINGW32__ */
+    if (endp > srcp+1 && isalpha(*srcp) && *(srcp+1) == ':') {
+        /* We first process the Evil Drive Letter */
+        Scm_DStringPutc(&buf, *srcp++);
+        Scm_DStringPutc(&buf, *srcp++);
+    }
+    if (endp > srcp && (*srcp == '/' || *srcp == '\\')) {
+        if (flags & SCM_PATH_CANONICALIZE) {
+            Scm_DStringPutc(&buf, SEPARATOR);
+            srcp = skip_separators(srcp, endp);
+        }
+    } else if (srcp == str) {
+        /* Path is relative (the srcp==str condition rules out the
+           drive letter case */
+        if (flags & SCM_PATH_ABSOLUTE) {
+            put_current_dir(&buf);
+        }
+    }
+    if (!(flags & SCM_PATH_CANONICALIZE)) {
+        copy_win32_path(&buf, srcp, endp);
+	return Scm_DStringGet(&buf, 0);
+    }
+#endif /* __MINGW32__ */
 
-    while (srcp < str+size) {
-        if (*srcp == '.') {
-            if (srcp == str+size-1) {
-                *dstp++ = '.'; /* preserve the last dot */
+    /* Canonicalization.  We used to have a tricky piece of code here
+       that avoids extra allocations, but have replaced it for
+       simple-minded code, since speed gain here won't contribute to
+       the overall performance anyway.  */
+    {
+        ScmObj comps = SCM_NIL; /* reverse list of components */
+        int cnt = 0;            /* # of components except ".."'s */
+        int final = FALSE;
+        int wentup = FALSE;     /* true if the last loop went up a dir */
+        const char *p;
+
+        for (;;) {
+            p = get_first_separator(srcp, endp);
+            if (p == NULL) {
+                final = TRUE;
+                p = endp;
+            }
+
+            if (p == srcp+1 && *srcp == '.') {
+                /* do nothing */
+            } else if (p == srcp+2 && srcp[0] == '.' && srcp[1] == '.') {
+                if (cnt > 0) {
+                    SCM_ASSERT(SCM_PAIRP(comps));
+                    comps = SCM_CDR(comps);
+                    cnt--;
+                    wentup = TRUE;
+                } else {
+                    comps = Scm_Cons(SCM_MAKE_STR(".."), comps);
+                    wentup = FALSE;
+                }
+            } else {
+                comps = Scm_Cons(Scm_MakeString(srcp, p-srcp, -1, 0), comps);
+                cnt++;
+                wentup = FALSE;
+            }
+            if (final) {
+                /* If we just went up a directory, we want to preserve the
+                   trailing separator in the result.  So we add an empty
+                   component. */
+                if (wentup) comps = Scm_Cons(SCM_MAKE_STR(""), comps);
                 break;
             }
-            if (*(srcp+1) == '/') {
-                srcp++;
-                SKIP_SLASH;
-                continue;
-            }
-            if (!bottomp
-                && *(srcp+1) == '.'
-                && (srcp == str+size-2 || *(srcp+2) == '/')) {
-
-                /* back up to parent dir */
-                char *q = dstp-2;
-                for (;q >= buf; q--) {
-                    if (*q == '/') break;
-                }
-                if (q >= buf) {
-                    dstp = q+1;
-                } else {
-                    bottomp = TRUE;
-                    *dstp++ = '.';
-                    *dstp++ = '.';
-                    *dstp++ = '/';
-                }
-                srcp += 3;
-                continue;
+            srcp = skip_separators(p, endp);
+        }
+        if (SCM_PAIRP(comps)) {
+            comps = Scm_ReverseX(comps);
+            Scm_DStringAdd(&buf, SCM_STRING(SCM_CAR(comps)));
+            SCM_FOR_EACH(comps, SCM_CDR(comps)) {
+                Scm_DStringPutc(&buf, SEPARATOR);
+                Scm_DStringAdd(&buf, SCM_STRING(SCM_CAR(comps)));
             }
         }
-        while ((*dstp++ = *srcp++) != '/' && srcp < str+size)
-            ;
-        SKIP_SLASH;
     }
-    *dstp = '\0';
-    return Scm_MakeString(buf, dstp-buf, -1, SCM_MAKSTR_COPYING);
+    return Scm_DStringGet(&buf, 0);        
 }
 
+/* Basename and dirname.
+   On Win32, we need to treat drive names specially, e.g.:
+   (sys-dirname "C:/a") == (sys-dirname "C:/") == (sys-dirname "C:") == "C:\\"
+   (sys-basename "C:/") == (sys-basename "C:) == "" 
+*/
+
 ScmObj Scm_BaseName(ScmString *filename)
 {
-    const char *p, *str = SCM_STRING_START(filename);
-    int i, size = SCM_STRING_SIZE(filename);
+    u_int size;
+    const char *path = Scm_GetStringContent(filename, &size, NULL, NULL);
+    const char *endp, *last;
+
+#if defined(__MINGW32__)
+    /* Ignore drive letter, for it can never be a part of basename. */
+    if (size >= 2 && path[1] == ':' && isalpha(path[0])) {
+        path += 2;
+        size -= 2;
+    }
+#endif /* __MINGW32__ */
 
     if (size == 0) return SCM_MAKE_STR("");
-    p = str+size-1;
-    while (*p == '/' && size > 0) { p--; size--; } /* ignore trailing '/' */
-    if (size == 0) return SCM_MAKE_STR("");
-    for (i = 0; i < size; i++, p--) {
-        if (*p == '/') break;
+    endp = truncate_trailing_separators(path, path+size);
+    last = get_last_separator(path, endp);
+    if (last == NULL) {
+        return Scm_MakeString(path, endp-path, -1, 0);
+    } else {
+        return Scm_MakeString(last+1, endp-last-1, -1, 0);
     }
-    return Scm_MakeString(p+1, i, -1, 0);
 }
 
 ScmObj Scm_DirName(ScmString *filename)
 {
-    const char *p, *str = SCM_STRING_START(filename);
-    int i, size = SCM_STRING_SIZE(filename);
-
-    if (size == 0) return SCM_MAKE_STR(".");
-    p = str+size-1;
-    while (*p == '/' && size > 0) { p--; size--; } /* ignore trailing '/' */
-    if (size == 0) return SCM_MAKE_STR("/");
-    for (i = size; i > 0; i--, p--) {
-        if (*p == '/') break;
+    u_int size;
+    const char *path = Scm_GetStringContent(filename, &size, NULL, NULL);
+    const char *endp, *last;
+#if defined(__MINGW32__)
+    int drive_letter = -1;
+    if (size >= 2 && path[1] == ':' && isalpha(path[0])) {
+        drive_letter = path[0];
+        path += 2;
+        size -= 2;
     }
-    if (i == 0) return SCM_MAKE_STR(".");
-    while (*p == '/' && i > 0) { p--; i--; } /* delete trailing '/' */
-    if (i == 0) return SCM_MAKE_STR("/");
-    return Scm_MakeString(str, i, -1, 0);
+#endif /* __MINGW32__ */
+
+    if (size == 0) { path = NULL; goto finale; }
+    endp = truncate_trailing_separators(path, path+size);
+    if (endp == path) { path = ROOTDIR, size = 1; goto finale; }
+    last = get_last_separator(path, endp);
+    if (last == NULL) { path = ".", size = 1; goto finale; }
+
+    /* we have "something/", and 'last' points to the last separator. */
+    last = truncate_trailing_separators(path, last);
+    if (last == path) {
+        path = ROOTDIR, size = 1;
+    } else {
+        size = last - path;
+    }
+ finale:
+#if defined(__MINGW32__)
+    if (drive_letter > 0) {
+        ScmObj z;
+        char p[3] = "x:";
+        p[0] = (char)drive_letter;
+        z = Scm_MakeString(p, 2, 2, SCM_MAKSTR_COPYING);
+        if (path) {
+            return Scm_StringAppendC(SCM_STRING(z), path, size, -1);
+        } else {
+            return Scm_StringAppendC(SCM_STRING(z), ROOTDIR, 1, -1);
+        }
+    }
+#endif /* __MINGW32__ */
+    if (path) return Scm_MakeString(path, size, -1, 0);
+    else      return Scm_MakeString(".", 1, 1, 0);
 }
 
+#undef ROOTDIR
+#undef SEPARATOR
+
 /* Make mkstemp() work even if the system doesn't have one. */
-ScmObj Scm_SysMkstemp(ScmString *template)
+int Scm_Mkstemp(char *templat)
 {
-#define MKSTEMP_PATH_MAX 1025  /* Geez, remove me */
-    char name[MKSTEMP_PATH_MAX];
-    ScmObj sname;
-    int siz = SCM_STRING_SIZE(template), fd;
-    if (siz >= MKSTEMP_PATH_MAX-6) { 
-	Scm_Error("pathname too long: %S", template);
-    }
-    memcpy(name, SCM_STRING_START(template), siz);
+    int fd = -1;
 #if defined(HAVE_MKSTEMP)
-    memcpy(name + siz, "XXXXXX", 6);
-    name[siz+6] = '\0';
-    SCM_SYSCALL(fd, mkstemp(name));
+    SCM_SYSCALL(fd, mkstemp(templat));
     if (fd < 0) Scm_SysError("mkstemp failed");
-    sname = SCM_MAKE_STR_COPYING(name);
-    SCM_RETURN(Scm_Values2(Scm_MakePortWithFd(sname, SCM_PORT_OUTPUT, fd,
-					      SCM_PORT_BUFFER_FULL, TRUE),
-			   sname));
+    return fd;
 #else   /*!defined(HAVE_MKSTEMP)*/
     /* Emulate mkstemp. */
+    int siz = strlen(templat);
+    if (siz < 6) {
+        Scm_Error("mkstemp - invalid template: %s", templat);
+    }
 #define MKSTEMP_MAX_TRIALS 65535   /* avoid infinite loop */
     {
 	u_long seed = (u_long)time(NULL);
-	int numtry;
+	int numtry, flags;
 	char suffix[7];
+#if defined(__MINGW32__)
+	flags = O_CREAT|O_EXCL|O_WRONLY|O_BINARY;
+#else  /* !__MINGW32__ */
+	flags = O_CREAT|O_EXCL|O_WRONLY;
+#endif /* !__MINGW32__ */
 	for (numtry=0; numtry<MKSTEMP_MAX_TRIALS; numtry++) {
 	    snprintf(suffix, 7, "%06x", seed&0xffffff);
-	    memcpy(name+siz, suffix, 7);
-#ifndef __MINGW32__
-	    SCM_SYSCALL(fd, open(name, O_CREAT|O_EXCL|O_WRONLY, 0600));
-#else  /*__MINGW32__*/
- 	    SCM_SYSCALL(fd, open(name, _O_CREAT|_O_EXCL|_O_WRONLY, 0600));
-#endif /*__MINGW32__*/
+	    memcpy(templat+siz-6, suffix, 7);
+	    SCM_SYSCALL(fd, open(templat, flags, 0600));
 	    if (fd >= 0) break;
 	    seed *= 2654435761UL;
 	}
 	if (numtry == MKSTEMP_MAX_TRIALS) {
 	    Scm_Error("mkstemp failed");
 	}
-	sname = SCM_MAKE_STR_COPYING(name);
-	SCM_RETURN(Scm_Values2(Scm_MakePortWithFd(sname, SCM_PORT_OUTPUT, fd,
-						  SCM_PORT_BUFFER_FULL, TRUE),
-			       sname));
     }
+    return fd;
 #endif /*!defined(HAVE_MKSTEMP)*/
 }
 
+
+ScmObj Scm_SysMkstemp(ScmString *templat)
+{
+#define MKSTEMP_PATH_MAX 1025  /* Geez, remove me */
+    char name[MKSTEMP_PATH_MAX];
+    ScmObj sname;
+    u_int siz;
+    int fd;
+    const char *t = Scm_GetStringContent(templat, &siz, NULL, NULL);
+    if (siz >= MKSTEMP_PATH_MAX-6) { 
+	Scm_Error("pathname too long: %S", templat);
+    }
+    memcpy(name, t, siz);
+    memcpy(name + siz, "XXXXXX", 6);
+    name[siz+6] = '\0';
+    fd = Scm_Mkstemp(name);
+    sname = SCM_MAKE_STR_COPYING(name);
+    SCM_RETURN(Scm_Values2(Scm_MakePortWithFd(sname, SCM_PORT_OUTPUT, fd,
+					      SCM_PORT_BUFFER_FULL, TRUE),
+			   sname));
+}
+
 /*===============================================================
  * Stat (sys/stat.h)
  */
@@ -528,8 +758,35 @@ static void time_print(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
     Scm_Printf(port, "#<%S %lu.%09lu>", t->type, t->sec, t->nsec);
 }
 
+static int time_compare(ScmObj x, ScmObj y, int equalp)
+{
+    ScmTime *tx = SCM_TIME(x);
+    ScmTime *ty = SCM_TIME(y);
+    
+    if (equalp) {
+        if (SCM_EQ(tx->type, ty->type)
+            && tx->sec == ty->sec
+            && tx->nsec == ty->nsec) {
+            return 0;
+        } else {
+            return 1;
+        }
+    } else {
+        if (!SCM_EQ(tx->type, ty->type)) {
+            Scm_Error("cannot compare different types of time objects: %S vs %S", x, y);
+        }
+        if (tx->sec < ty->sec) return -1;
+        if (tx->sec == ty->sec) {
+            if (tx->nsec < ty->nsec) return -1;
+            if (tx->nsec == ty->nsec) return 0;
+            else return 1;
+        } 
+        else return 1;
+    }
+}
+
 SCM_DEFINE_BUILTIN_CLASS(Scm_TimeClass,
-                         time_print, NULL, NULL,
+                         time_print, time_compare, NULL,
                          time_allocate, SCM_CLASS_DEFAULT_CPL);
 
 ScmObj Scm_MakeTime(ScmObj type, long sec, long nsec)
@@ -541,17 +798,33 @@ ScmObj Scm_MakeTime(ScmObj type, long sec, long nsec)
     return SCM_OBJ(t);
 }
 
-ScmObj Scm_CurrentTime(void)
+/* Abstract gettimeofday() */
+void Scm_GetTimeOfDay(u_long *sec, u_long *usec)
 {
-#ifdef HAVE_GETTIMEOFDAY
+#if defined(HAVE_GETTIMEOFDAY)
     struct timeval tv;
     int r;
     SCM_SYSCALL(r, gettimeofday(&tv, NULL));
     if (r < 0) Scm_SysError("gettimeofday failed");
-    return Scm_MakeTime(SCM_SYM_TIME_UTC, (long)tv.tv_sec, (long)tv.tv_usec*1000);
-#else  /* !HAVE_GETTIMEOFDAY */
-    return Scm_MakeTime(SCM_SYM_TIME_UTC, (long)time(NULL), 0);
-#endif /* !HAVE_GETTIMEOFDAY */
+    *sec = (u_long)tv.tv_sec;
+    *usec = (u_long)tv.tv_usec;
+#elif defined(__MINGW32__)
+    FILETIME ft;
+    GetSystemTimeAsFileTime(&ft);
+    SCM_FILETIME_TO_UNIXTIME(ft, *sec, *usec);
+#else  /* !HAVE_GETTIMEOFDAY && !__MINGW32 */
+    /* Last resort */
+    *sec = (u_long)time(NULL);
+    *usec = 0;
+#endif /* !HAVE_GETTIMEOFDAY && !__MINGW32 */
+}
+
+
+ScmObj Scm_CurrentTime(void)
+{
+    u_long sec, usec;
+    Scm_GetTimeOfDay(&sec, &usec);
+    return Scm_MakeTime(SCM_SYM_TIME_UTC, sec, usec*1000);
 }
 
 ScmObj Scm_IntSecondsToTime(long sec)
@@ -607,7 +880,7 @@ static void time_nsec_set(ScmTime *t, ScmObj val)
         Scm_Error("real number required, but got %S", val);
     }
     if ((l = Scm_GetInteger(val)) >= 1000000000) {
-        Scm_Error("nanoseconds out of range: %l", l);
+        Scm_Error("nanoseconds out of range: %ld", l);
     }
     t->nsec = l;
 }
@@ -762,6 +1035,35 @@ static ScmClassStaticSlotSpec tm_slots[] = {
     { NULL }
 };
 
+/*===============================================================
+ * Yielding CPU (sched.h, if available)
+ */
+
+/* If sched_yield is not available, we make the calling thread sleep
+   small amount of time, hoping there are other threads that can run
+   in place. */
+void
+Scm_YieldCPU(void)
+{
+#if defined(HAVE_SCHED_YIELD)
+    sched_yield();
+#elif defined(HAVE_NANOSLEEP)
+    struct timespec spec;
+    spec.tv_sec = 0;
+    spec.tv_nsec = 1;
+    nanosleep(&spec, NULL);
+#elif defined(HAVE_SELECT)
+    struct timeval tv;
+    tv.tv_sec = 0;
+    tv.tv_usec = 1;
+    select(0, NULL, NULL, NULL, &tv);
+#elif defined(__MINGW32__)
+    Sleep(10);
+#else /* the last resort */
+    sleep(1);
+#endif
+}
+
 /*===============================================================
  * Groups (grp.h)
  */
@@ -776,23 +1078,18 @@ SCM_DEFINE_BUILTIN_CLASS_SIMPLE(Scm_SysGroupClass, grp_print);
 
 static ScmObj make_group(struct group *g)
 {
-    ScmObj head = SCM_NIL, tail = SCM_NIL, p;
-    char **memp;
     ScmSysGroup *sg = SCM_NEW(ScmSysGroup);
     SCM_SET_CLASS(sg, SCM_CLASS_SYS_GROUP);
     
     sg->name = SCM_MAKE_STR_COPYING(g->gr_name);
-#ifdef HAVE_GR_PASSWD
+#ifdef HAVE_STRUCT_GROUP_GR_PASSWD
     sg->passwd = SCM_MAKE_STR_COPYING(g->gr_passwd);
 #else
     sg->passwd = SCM_FALSE;
 #endif
     sg->gid = Scm_MakeInteger(g->gr_gid);
-    for (memp = g->gr_mem; *memp; memp++) {
-        p = SCM_MAKE_STR_COPYING(*memp);
-        SCM_APPEND1(head, tail, p);
-    }
-    sg->mem = head;
+    sg->mem = Scm_CStringArrayToList((const char**)g->gr_mem, -1,
+                                     SCM_MAKSTR_COPYING);
     return SCM_OBJ(sg);
 }
 
@@ -858,17 +1155,17 @@ static ScmObj make_passwd(struct passwd *pw)
     sp->name = SCM_MAKE_STR_COPYING(pw->pw_name);
     sp->uid = Scm_MakeInteger(pw->pw_uid);
     sp->gid = Scm_MakeInteger(pw->pw_gid);
-#ifdef HAVE_PW_PASSWD
+#ifdef HAVE_STRUCT_PASSWD_PW_PASSWD
     sp->passwd = SCM_MAKE_STR_COPYING(pw->pw_passwd);
 #else
     sp->passwd = SCM_FALSE;
 #endif
-#ifdef HAVE_PW_GECOS
+#ifdef HAVE_STRUCT_PASSWD_PW_GECOS
     sp->gecos = SCM_MAKE_STR_COPYING(pw->pw_gecos);
 #else
     sp->gecos = SCM_FALSE;
 #endif
-#ifdef HAVE_PW_CLASS
+#ifdef HAVE_STRUCT_PASSWD_PW_CLASS
     sp->pwclass = SCM_MAKE_STR_COPYING(pw->pw_class);
 #else
     sp->pwclass = SCM_FALSE;
@@ -962,42 +1259,96 @@ int Scm_IsSugid(void)
  *   reasonable way.  For now, iomap is ignored.
  */
 
-ScmObj Scm_SysExec(ScmString *file, ScmObj args, ScmObj iomap, int forkp)
+ScmObj Scm_SysExec(ScmString *file, ScmObj args, ScmObj iomap,
+                   ScmSysSigset *mask, int flags)
 {
-    int argc = Scm_Length(args), i, j, maxfd, iollen;
-    int *tofd = NULL, *fromfd = NULL, *tmpfd = NULL;
+    int argc = Scm_Length(args), *fds;
     char **argv;
     const char *program;
-    ScmObj ap, iop;
-    pid_t pid;
+    pid_t pid = 0;
+    int forkp = flags & SCM_EXEC_WITH_FORK;
 
     if (argc < 1) {
         Scm_Error("argument list must have at least one element: %S", args);
     }
 
     /* make a C array of C strings */    
-    argv = SCM_NEW_ARRAY(char *, argc+1);
-    for (i=0, ap = args; i<argc; i++, ap = SCM_CDR(ap)) {
-        if (!SCM_STRINGP(SCM_CAR(ap)))
-            Scm_Error("bad argument (string required): %S", SCM_CAR(ap));
-        argv[i] = Scm_GetString(SCM_STRING(SCM_CAR(ap)));
-    }
-    argv[i] = NULL;
+    argv = Scm_ListToCStringArray(args, TRUE, NULL);
     program = Scm_GetStringConst(file);
 
 #ifndef __MINGW32__
     /* setting up iomap table */
-    iollen = Scm_Length(iomap);
+    fds = Scm_SysPrepareFdMap(iomap);
+    
+    /* When requested, call fork() here. */
+    if (forkp) {
+        SCM_SYSCALL(pid, fork());
+        if (pid < 0) Scm_SysError("fork failed");
+    }
+
+    /* Now we swap file descriptors and exec(). */
+    if (!forkp || pid == 0) {
+        Scm_SysSwapFds(fds);
+        if (mask) {
+            Scm_ResetSignalHandlers(&mask->set);
+            Scm_SysSigmask(SIG_SETMASK, mask);
+        }
+        execvp(program, (char *const*)argv);
+        /* here, we failed */
+        Scm_Panic("exec failed: %s: %s", program, strerror(errno));
+    }
+
+    /* We come here only when fork is requested. */
+    return Scm_MakeInteger(pid);
+#else  /* __MINGW32__ */
+    if (forkp) {
+	Scm_Error("fork() not supported on MinGW port");
+    } else {
+	execvp(program, (const char *const*)argv);
+	Scm_Panic("exec failed: %s: %s", program, strerror(errno));	
+    }
+    return SCM_FALSE; /* dummy */
+#endif /* __MINGW32__ */
+}
+
+/* Two auxiliary functions to support iomap feature.  They are exposed
+   so that the library can implement iomap feature as the same way as
+   sys-exec.
+
+   The first function, Scm_SysPrepareFdMap, walks iomap structure and
+   prepare a table of file descriptors to modify.  The second function,
+   Scm_SysSwapFds, takes the table and modifies process's file descriptors.
+
+   We need to split this feature to two function, since it is unsafe
+   to raise an error after fork() in multi-threaded environment.
+   Scm_SysPrepareFdMap may throw an error if passed iomap contains
+   invalid entries.  On the other hand, Scm_SysSwapFds just aborts if
+   things goes wrong---not only because of the MT-safety issue, but also
+   it is generally impossible to handle errors reasonably since we don't
+   even sure we have stdios.   And the client code is supposed to call  
+   fork() between these functions.
+
+   The client code should treat the returned pointer of Scm_SysPrepareFdMap
+   opaque, and pass it to Scm_SysSwapFds as is.
+*/
+int *Scm_SysPrepareFdMap(ScmObj iomap)
+{
+    int *fds = NULL;
+#ifndef __MINGW32__
     if (SCM_PAIRP(iomap)) {
+        ScmObj iop;
+        int iollen = Scm_Length(iomap), i = 0;
+        int *tofd, *fromfd;
+        
         /* check argument vailidity before duping file descriptors, so that
            we can still use Scm_Error */
         if (iollen < 0) {
             Scm_Error("proper list required for iolist, but got %S", iomap);
         }
-        tofd   = SCM_NEW_ATOMIC2(int *, iollen * sizeof(int));
-        fromfd = SCM_NEW_ATOMIC2(int *, iollen * sizeof(int));
-        tmpfd  = SCM_NEW_ATOMIC2(int *, iollen * sizeof(int));
-        i = 0;
+        fds    = SCM_NEW_ATOMIC2(int *, 2 * iollen * sizeof(int) + 1);
+        fds[0] = iollen;
+        tofd   = fds + 1;
+        fromfd = fds + 1 + iollen;
         SCM_FOR_EACH(iop, iomap) {
             ScmObj port, elt = SCM_CAR(iop);
             if (!SCM_PAIRP(elt) || !SCM_INTP(SCM_CAR(elt))
@@ -1027,57 +1378,50 @@ ScmObj Scm_SysExec(ScmString *file, ScmObj args, ScmObj iomap, int forkp)
             i++;
         }
     }
+#endif /* __MINGW32__ */
+    return fds;
+}
+
+void Scm_SysSwapFds(int *fds)
+{
+#ifndef __MINGW32__
+    int *tofd, *fromfd, nfds, maxfd, i, j, fd;
     
-    /* When requested, call fork() here. */
-    if (forkp) {
-        SCM_SYSCALL(pid, fork());
-        if (pid < 0) Scm_SysError("fork failed");
+    if (fds == NULL) return;
+
+    nfds   = fds[0];
+    tofd   = fds + 1;
+    fromfd = fds + 1 + nfds;
+
+    /* TODO: use getdtablehi if available */
+    if ((maxfd = sysconf(_SC_OPEN_MAX)) < 0) {
+        Scm_Panic("failed to get OPEN_MAX value from sysconf");
     }
 
-    /* Now we swap file descriptors and exec().
-       We can't throw an error anymore! */
-    if (!forkp || pid == 0) {
-        /* TODO: use getdtablehi if available */
-        if ((maxfd = sysconf(_SC_OPEN_MAX)) < 0) {
-            Scm_Panic("failed to get OPEN_MAX value from sysconf");
-        }
-
-        for (i=0; i<iollen; i++) {
-            if (tofd[i] == fromfd[i]) continue;
-            for (j=i+1; j<iollen; j++) {
-                if (tofd[i] == fromfd[j]) {
-                    int tmp = dup(tofd[i]);
-                    if (tmp < 0) Scm_Panic("dup failed: %s", strerror(errno));
-                    fromfd[j] = tmp;
-                }
+    /* Dup fromfd to the corresponding tofd.  We need to be careful
+       not to override the destination fd if it will be used. */
+    for (i=0; i<nfds; i++) {
+        if (tofd[i] == fromfd[i]) continue;
+        for (j=i+1; j<nfds; j++) {
+            if (tofd[i] == fromfd[j]) {
+                int tmp = dup(tofd[i]);
+                if (tmp < 0) Scm_Panic("dup failed: %s", strerror(errno));
+                fromfd[j] = tmp;
             }
-            if (dup2(fromfd[i], tofd[i]) < 0)
-                Scm_Panic("dup2 failed: %s", strerror(errno));
         }
-        for (i=0; i<maxfd; i++) {
-            for (j=0; j<iollen; j++) {
-                if (i == tofd[j]) break;
-            }
-            if (j == iollen) close(i);
-        }
-        execvp(program, (char *const*)argv);
-        /* here, we failed */
-        Scm_Panic("exec failed: %s: %s", program, strerror(errno));
+        if (dup2(fromfd[i], tofd[i]) < 0)
+            Scm_Panic("dup2 failed: %s", strerror(errno));
     }
-
-    /* We come here only when fork is requested. */
-    return Scm_MakeInteger(pid);
-#else  /* __MINGW32__ */
-    if (forkp) {
-	Scm_Error("fork() not supported on MinGW port");
-    } else {
-	execvp(program, (const char *const*)argv);
-	Scm_Panic("exec failed: %s: %s", program, strerror(errno));	
+    
+    /* Close unused fds */
+    for (fd=0; fd<maxfd; fd++) {
+        for (j=0; j<nfds; j++) if (fd == tofd[j]) break;
+        if (j == nfds) close(fd);
     }
-    return SCM_FALSE; /* dummy */
 #endif /* __MINGW32__ */
 }
 
+
 /*===============================================================
  * select
  */
@@ -1423,12 +1767,12 @@ int link(const char *existing, const char *newpath)
 void Scm__InitSystem(void)
 {
     ScmModule *mod = Scm_GaucheModule();
-    Scm_InitBuiltinClass(&Scm_SysStatClass, "<sys-stat>", stat_slots, 0, mod);
-    Scm_InitBuiltinClass(&Scm_TimeClass, "<time>", time_slots, 0, mod);
-    Scm_InitBuiltinClass(&Scm_SysTmClass, "<sys-tm>", tm_slots, 0, mod);
-    Scm_InitBuiltinClass(&Scm_SysGroupClass, "<sys-group>", grp_slots, 0, mod);
-    Scm_InitBuiltinClass(&Scm_SysPasswdClass, "<sys-passwd>", pwd_slots, 0, mod);
+    Scm_InitStaticClass(&Scm_SysStatClass, "<sys-stat>", mod, stat_slots, 0);
+    Scm_InitStaticClass(&Scm_TimeClass, "<time>", mod, time_slots, 0);
+    Scm_InitStaticClass(&Scm_SysTmClass, "<sys-tm>", mod, tm_slots, 0);
+    Scm_InitStaticClass(&Scm_SysGroupClass, "<sys-group>", mod, grp_slots, 0);
+    Scm_InitStaticClass(&Scm_SysPasswdClass, "<sys-passwd>", mod, pwd_slots, 0);
 #ifdef HAVE_SELECT
-    Scm_InitBuiltinClass(&Scm_SysFdsetClass, "<sys-fdset>", NULL, 0, mod);
+    Scm_InitStaticClass(&Scm_SysFdsetClass, "<sys-fdset>", mod, NULL, 0);
 #endif
 }
