@@ -1,7 +1,7 @@
 /*
  * write.c - writer
  *
- *   Copyright (c) 2000-2004 Shiro Kawai, All rights reserved.
+ *   Copyright (c) 2000-2006 Shiro Kawai, All rights reserved.
  * 
  *   Redistribution and use in source and binary forms, with or without
  *   modification, are permitted provided that the following conditions
@@ -30,7 +30,7 @@
  *   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
  *   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  *
- *  $Id: write.c,v 1.46 2004/07/19 09:19:07 shirok Exp $
+ *  $Id: write.c,v 1.64 2006/11/04 09:57:00 shirok Exp $
  */
 
 #include <stdio.h>
@@ -45,6 +45,8 @@ static void write_ss(ScmObj obj, ScmPort *port, ScmWriteContext *ctx);
 static void write_ss_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx);
 static void write_object(ScmObj obj, ScmPort *out, ScmWriteContext *ctx);
 static ScmObj write_object_fallback(ScmObj *args, int nargs, ScmGeneric *gf);
+static void format_write(ScmObj obj, ScmPort *port, ScmWriteContext *ctx,
+                         int sharedp);
 SCM_DEFINE_GENERIC(Scm_GenericWriteObject, write_object_fallback, NULL);
 
 /*============================================================
@@ -141,13 +143,13 @@ void Scm_Write(ScmObj obj, ScmObj p, int mode)
 
     /* if this is a "walk" pass of write/ss, dispatch to the walker */
     if (port->flags & SCM_PORT_WALKING) {
-        SCM_ASSERT(SCM_PAIRP(port->data)&&SCM_HASHTABLEP(SCM_CDR(port->data)));
+        SCM_ASSERT(SCM_PAIRP(port->data)&&SCM_HASH_TABLE_P(SCM_CDR(port->data)));
         write_walk(obj, port, &ctx);
         return;
     }
     /* if this is a "output" pass of write/ss, call the recursive routine */
     if (port->flags & SCM_PORT_WRITESS) {
-        SCM_ASSERT(SCM_PAIRP(port->data)&&SCM_HASHTABLEP(SCM_CDR(port->data)));
+        SCM_ASSERT(SCM_PAIRP(port->data)&&SCM_HASH_TABLE_P(SCM_CDR(port->data)));
         write_ss_rec(obj, port, &ctx);
         return;
     }
@@ -179,18 +181,26 @@ int Scm_WriteLimited(ScmObj obj, ScmObj port, int mode, int width)
 {
     ScmWriteContext ctx;
     ScmObj out;
-    int nc;
+    int nc, sharedp = FALSE;
     
     if (!SCM_OPORTP(port))
         Scm_Error("output port required, but got %S", port);
     out = Scm_MakeOutputStringPort(TRUE);
+    SCM_PORT(out)->data = SCM_PORT(port)->data;
     ctx.mode = mode;
     ctx.flags = WRITE_LIMITED;
     ctx.limit = width;
     /* if case mode is not specified, use default taken from VM default */
     if (SCM_WRITE_CASE(&ctx) == 0) ctx.mode |= DEFAULT_CASE;
+    /* the walk pass does not produce any output, so we return immediately. */
+    if (SCM_PORT(port)->flags & SCM_PORT_WALKING) {
+        SCM_ASSERT(SCM_PAIRP(SCM_PORT(port)->data)&&SCM_HASH_TABLE_P(SCM_CDR(SCM_PORT(port)->data)));
+        write_walk(obj, SCM_PORT(port), &ctx);
+        return 0;               /* doesn't really matter */
+    }
     /* we don't need to lock out, for it is private. */
-    write_ss_rec(obj, SCM_PORT(out), &ctx);
+    sharedp = SCM_WRITE_MODE(&ctx) == SCM_WRITE_SHARED;
+    format_write(obj, SCM_PORT(out), &ctx, sharedp);
     nc = outlen(SCM_PORT(out));
     if (nc > width) {
         ScmObj sub = Scm_Substring(SCM_STRING(Scm_GetOutputString(SCM_PORT(out))),
@@ -209,6 +219,7 @@ int Scm_WriteLimited(ScmObj obj, ScmObj port, int mode, int width)
 
 int Scm_WriteCircular(ScmObj obj, ScmObj port, int mode, int width)
 {
+    ScmObj out;
     ScmWriteContext ctx;
     int nc;
 
@@ -223,30 +234,37 @@ int Scm_WriteCircular(ScmObj obj, ScmObj port, int mode, int width)
         ctx.limit = width;
     }
     ctx.ncirc = 0;
-    ctx.table = SCM_HASHTABLE(Scm_MakeHashTable(SCM_HASH_ADDRESS, NULL, 8));
+    ctx.table = SCM_HASH_TABLE(Scm_MakeHashTableSimple(SCM_HASH_EQ, 8));
 
-    if (width > 0) {
-        ScmObj out = Scm_MakeOutputStringPort(TRUE);
-        /* no need to lock out, for it is private */
-        write_ss(obj, SCM_PORT(out), &ctx);
-        nc = outlen(SCM_PORT(out));
-        if (nc > width) {
-            ScmObj sub = Scm_Substring(SCM_STRING(Scm_GetOutputString(SCM_PORT(out))),
-                                       0, width);
-            SCM_PUTS(sub, port); /* this locks port */
-            return -1;
-        } else {
-            SCM_PUTS(Scm_GetOutputString(SCM_PORT(out)), port); /* this locks port */
-            return nc;
-        }
-    } else {
+    if (width <= 0) {
         ScmVM *vm = Scm_VM();
         PORT_LOCK(SCM_PORT(port), vm);
         PORT_SAFE_CALL(SCM_PORT(port),
-                       write_ss(obj, SCM_PORT(port), &ctx));
+                       format_write(obj, SCM_PORT(port), &ctx, TRUE));
         PORT_UNLOCK(SCM_PORT(port));
+        return 0;
+    }
+
+    if (SCM_PORT(port)->flags & SCM_PORT_WALKING) {
+        SCM_ASSERT(SCM_PAIRP(SCM_PORT(port)->data)&&SCM_HASH_TABLE_P(SCM_CDR(SCM_PORT(port)->data)));
+        write_walk(obj, SCM_PORT(port), &ctx);
+        return 0;               /* doesn't really matter */
+    }
+
+    out = Scm_MakeOutputStringPort(TRUE);
+    SCM_PORT(out)->data = SCM_PORT(port)->data;
+    /* no need to lock out, for it is private */
+    format_write(obj, SCM_PORT(out), &ctx, TRUE);
+    nc = outlen(SCM_PORT(out));
+    if (nc > width) {
+        ScmObj sub = Scm_Substring(SCM_STRING(Scm_GetOutputString(SCM_PORT(out))),
+                                   0, width);
+        SCM_PUTS(sub, port); /* this locks port */
+        return -1;
+    } else {
+        SCM_PUTS(Scm_GetOutputString(SCM_PORT(out)), port); /* this locks port */
+        return nc;
     }
-    return 0;
 }
 
 /*===================================================================
@@ -279,7 +297,8 @@ static void write_general(ScmObj obj, ScmPort *out, ScmWriteContext *ctx)
    may not return to VM immediately. */
 static void write_object(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
 {
-    Scm_Apply(SCM_OBJ(&Scm_GenericWriteObject), SCM_LIST2(obj, SCM_OBJ(port)));
+    Scm_ApplyRec(SCM_OBJ(&Scm_GenericWriteObject),
+                 SCM_LIST2(obj, SCM_OBJ(port)));
 }
 
 /* Default method for write-object */
@@ -291,7 +310,10 @@ static ScmObj write_object_fallback(ScmObj *args, int nargs, ScmGeneric *gf)
                   Scm_ArrayToList(args, nargs));
     }
     klass = Scm_ClassOf(args[0]);
-    Scm_Printf(SCM_PORT(args[1]), "#<%A %p>", klass->name, args[0]);
+    Scm_Printf(SCM_PORT(args[1]), "#<%A%s%p>",
+               klass->name,
+               (SCM_FALSEP(klass->redefined)? " " : ":redefined "),
+               args[0]);
     return SCM_TRUE;
 }
 
@@ -318,7 +340,7 @@ static ScmObj write_object_fallback(ScmObj *args, int nargs, ScmGeneric *gf)
 static ScmPortVTable walker_port_vtable = {
     NULL, NULL, NULL, NULL, NULL,
     NULL, NULL, NULL, NULL, NULL,
-    NULL, NULL, NULL
+    NULL, NULL
 };
 
 static ScmPort *make_walker_port(void)
@@ -326,8 +348,9 @@ static ScmPort *make_walker_port(void)
     ScmPort *port;
     ScmObj ht;
                                           
-    port = SCM_PORT(Scm_MakeVirtualPort(SCM_PORT_OUTPUT, &walker_port_vtable));
-    ht = Scm_MakeHashTable(SCM_HASH_ADDRESS, NULL, 0);
+    port = SCM_PORT(Scm_MakeVirtualPort(SCM_CLASS_PORT, SCM_PORT_OUTPUT,
+                                        &walker_port_vtable));
+    ht = Scm_MakeHashTableSimple(SCM_HASH_EQ, 0);
     port->data = Scm_Cons(SCM_MAKE_INT(0), ht);
     port->flags = SCM_PORT_WALKING;
     return port;
@@ -340,7 +363,7 @@ static void write_walk(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
     ScmHashTable *ht;
     ScmObj elt;
     
-    ht = SCM_HASHTABLE(SCM_CDR(port->data));
+    ht = SCM_HASH_TABLE(SCM_CDR(port->data));
 
     for (;;) {
         if (!SCM_PTRP(obj) || SCM_SYMBOLP(obj) || SCM_KEYWORDP(obj)
@@ -358,7 +381,7 @@ static void write_walk(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
             obj = SCM_CDR(obj);
             continue;
         }
-        if (SCM_STRINGP(obj) && SCM_STRING_SIZE(obj) > 0) {
+        if (SCM_STRINGP(obj) && !SCM_STRING_NULL_P(obj)) {
             e = Scm_HashTableGet(ht, obj);
             if (e) { e->value = SCM_TRUE; return; }
             Scm_HashTablePut(ht, obj, SCM_FALSE);
@@ -380,7 +403,6 @@ static void write_walk(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
         else {
             /* Now we have user-defined object.
                Call the user's print routine. */
-            ScmClass *c;
             e = Scm_HashTableGet(ht, obj);
             if (e) { e->value = SCM_TRUE; return; }
             Scm_HashTablePut(ht, obj, SCM_FALSE);
@@ -402,8 +424,8 @@ static void write_ss_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
         if (outlen(port) >= ctx->limit) return;
     }
 
-    if (SCM_PAIRP(port->data) && SCM_HASHTABLEP(SCM_CDR(port->data))) {
-        ht = SCM_HASHTABLE(SCM_CDR(port->data));
+    if (SCM_PAIRP(port->data) && SCM_HASH_TABLE_P(SCM_CDR(port->data))) {
+        ht = SCM_HASH_TABLE(SCM_CDR(port->data));
     }
 
     if (!SCM_PTRP(obj)) {
@@ -435,9 +457,6 @@ static void write_ss_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
                 else                  Scm_PutcUnsafe(ch, port);
             }
         }
-        else if (SCM_VM_INSNP(obj)) {
-            Scm__VMInsnWrite(obj, port, ctx);
-        }
         else Scm_Panic("write: got a bogus object: %08x", SCM_WORD(obj));
         return;
     }
@@ -447,7 +466,7 @@ static void write_ss_rec(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
         return;
     }
     
-    if ((SCM_STRINGP(obj) && SCM_STRING_SIZE(obj) == 0)
+    if ((SCM_STRINGP(obj) && SCM_STRING_NULL_P(obj))
         || (SCM_VECTORP(obj) && SCM_VECTOR_SIZE(obj) == 0)) {
         /* special case where we don't put a reference tag. */
         write_general(obj, port, ctx);
@@ -547,6 +566,7 @@ static void write_ss(ScmObj obj, ScmPort *port, ScmWriteContext *ctx)
     
     /* pass 1 */
     write_walk(obj, walker_port, ctx);
+    Scm_ClosePort(walker_port);
 
     /* pass 2 */
     /* TODO: we need to rewind port mode */
@@ -580,12 +600,12 @@ static void format_write(ScmObj obj, ScmPort *port, ScmWriteContext *ctx,
                          int sharedp)
 {
     if (port->flags & SCM_PORT_WALKING) {
-        SCM_ASSERT(SCM_PAIRP(port->data)&&SCM_HASHTABLEP(SCM_CDR(port->data)));
+        SCM_ASSERT(SCM_PAIRP(port->data)&&SCM_HASH_TABLE_P(SCM_CDR(port->data)));
         write_walk(obj, port, ctx);
         return;
     }
     if (port->flags & SCM_PORT_WRITESS) {
-        SCM_ASSERT(SCM_PAIRP(port->data)&&SCM_HASHTABLEP(SCM_CDR(port->data)));
+        SCM_ASSERT(SCM_PAIRP(port->data)&&SCM_HASH_TABLE_P(SCM_CDR(port->data)));
         write_ss_rec(obj, port, ctx);
         return;
     }
@@ -601,7 +621,7 @@ static void format_pad(ScmPort *out, ScmString *str,
                        int mincol, int colinc, ScmChar padchar,
                        int rightalign)
 {
-    int padcount = mincol- SCM_STRING_LENGTH(str);
+    int padcount = mincol- SCM_STRING_BODY_LENGTH(SCM_STRING_BODY(str));
     int i;
     
     if (padcount > 0) {
@@ -650,7 +670,7 @@ static void format_sexp(ScmPort *out, ScmObj arg,
     tmpstr = SCM_STRING(Scm_GetOutputString(SCM_PORT(tmpout)));
 
     if (maxcol > 0 && nwritten < 0) {
-        const char *s = SCM_STRING_START(tmpstr), *e;
+        const char *s = Scm_GetStringContent(tmpstr, NULL, NULL, NULL), *e;
         if (dots && maxcol > 4) {
             e = Scm_StringPosition(tmpstr, maxcol-4);
             Scm_PutzUnsafe(s, e-s, out);
@@ -686,7 +706,7 @@ static void format_integer(ScmPort *out, ScmObj arg,
     if (nparams>2 && SCM_CHARP(params[2])) commachar = SCM_CHAR_VALUE(params[2]);
     if (nparams>3 && SCM_INTP(params[3])) commainterval = SCM_INT_VALUE(params[3]);
     str = Scm_NumberToString(arg, radix, use_upper);
-    if (alwayssign && SCM_STRING_START(str)[0] != '-') {
+    if (alwayssign && SCM_STRING_BODY_START(SCM_STRING_BODY(str))[0] != '-') {
         str = Scm_StringAppend2(SCM_STRING(SCM_MAKE_STR("+")),
                                 SCM_STRING(str));
     }
@@ -697,8 +717,9 @@ static void format_integer(ScmPort *out, ScmObj arg,
            contradicts its examples; it is ambiguous about what happens
            if the number is padded. */
         ScmDString tmpout;
-        const char *ptr = SCM_STRING_START(str);
-        int num_digits = SCM_STRING_LENGTH(str), colcnt;
+        u_int num_digits, colcnt;
+        const char *ptr = Scm_GetStringContent(SCM_STRING(str), &num_digits,
+                                               NULL, NULL);
 
         Scm_DStringInit(&tmpout);
         if (*ptr == '-' || *ptr == '+') {
@@ -713,7 +734,7 @@ static void format_integer(ScmPort *out, ScmObj arg,
             Scm_DStringPutz(&tmpout, ptr+colcnt, commainterval);
             colcnt += commainterval;
         }
-        str = Scm_DStringGet(&tmpout);
+        str = Scm_DStringGet(&tmpout, 0);
     }
     format_pad(out, SCM_STRING(str), mincol, 1, padchar, TRUE);
 }
@@ -767,8 +788,8 @@ static void format_proc(ScmPort *out, ScmString *fmt, ScmObj args, int sharedp)
                 if (numParams == 0) {
                     format_write(arg, out, &sctx, sharedp);
                 } else {
-                    format_sexp(out, arg, params, numParams, atflag,
-                                colonflag, SCM_WRITE_WRITE);
+                    format_sexp(out, arg, params, numParams, atflag, colonflag,
+                                sharedp? SCM_WRITE_SHARED:SCM_WRITE_WRITE);
                 }
                 break;
             case 'a':; case 'A':;
@@ -777,8 +798,11 @@ static void format_proc(ScmPort *out, ScmString *fmt, ScmObj args, int sharedp)
                     /* short path */
                     format_write(arg, out, &actx, sharedp);
                 } else {
-                    format_sexp(out, arg, params, numParams, atflag,
-                                colonflag, SCM_WRITE_DISPLAY);
+                    /* TODO: Here we discard sharedp flag, since we don't yet
+                       have 'display/ss' mode.  We should separate
+                       write/display distinction and shared-write mode. */
+                    format_sexp(out, arg, params, numParams, atflag, colonflag,
+                                SCM_WRITE_DISPLAY);
                 }
                 break;
             case 'd':; case 'D':;
@@ -854,7 +878,7 @@ static void format_proc(ScmPort *out, ScmString *fmt, ScmObj args, int sharedp)
                         Scm_Error("'~*' format directive refers outside of argument list in %S", fmt);
                     }
                     argcnt = argindex;
-                    args = Scm_ListTail(oargs, argcnt);
+                    args = Scm_ListTail(oargs, argcnt, SCM_UNBOUND);
                     break;
                 }
             case 'v':; case 'V':;
@@ -1060,7 +1084,9 @@ static void vprintf_proc(ScmPort *out, const char *fmt, ScmObj args,
                     /* TODO: support right adjustment such as %-10s.
                        Currently we ignore minus sign and pad chars
                        on the right. */
-                    for (len = SCM_STRING_LENGTH(val); len < width; len++) {
+                    for (len = SCM_STRING_BODY_LENGTH(SCM_STRING_BODY(val));
+                         len < width;
+                         len++) {
                         Scm_PutcUnsafe(' ', out);
                     }
                     break;
@@ -1149,7 +1175,7 @@ static void vprintf_proc(ScmPort *out, const char *fmt, ScmObj args,
             break;
         }
         if (c == 0) {
-            Scm_Error("incomplete %-directive in format string: %s", fmt);
+            Scm_Error("incomplete %%-directive in format string: %s", fmt);
         }
     }
 }
@@ -1180,7 +1206,7 @@ void Scm_Vprintf(ScmPort *out, const char *fmt, va_list ap, int sharedp)
             case 'o':; case 'u':; case 'x':; case 'X':
                 {
                     unsigned long val = va_arg(ap, unsigned long);
-                    SCM_APPEND1(h, t, Scm_MakeIntegerFromUI(val));
+                    SCM_APPEND1(h, t, Scm_MakeIntegerU(val));
                     break;
                 }
             case 'e':; case 'E':; case 'f':; case 'g':; case 'G':
@@ -1204,7 +1230,7 @@ void Scm_Vprintf(ScmPort *out, const char *fmt, va_list ap, int sharedp)
             case 'p':
                 {
                     void *val = va_arg(ap, void *);
-                    SCM_APPEND1(h, t, Scm_MakeIntegerFromUI((unsigned long)val));
+                    SCM_APPEND1(h, t, Scm_MakeIntegerU((unsigned long)val));
                     break;
                 }
             case 'S':; case 'A':
@@ -1225,7 +1251,7 @@ void Scm_Vprintf(ScmPort *out, const char *fmt, va_list ap, int sharedp)
             break;
         }
         if (c == 0) {
-            Scm_Error("incomplete %-directive in format string: %s", fmt);
+            Scm_Error("incomplete %%-directive in format string: %s", fmt);
         }
     }
     /*
